# -*- coding: utf-8 -*-
import logging
import urllib.request
import zipfile
from pathlib import Path

from eyepy.core.base import Oct

logger = logging.getLogger(__name__)

SAMPLE_DATA = {
    "drusen_patient": (
        "https://uni-bonn.sciebo.de/s/VD8CPAgDKp2EYlm/download",
        Oct.from_heyex_xml,
    )
}

EYEPY_DATA_DIR = Path("~/.eyepy/data").expanduser()
if not EYEPY_DATA_DIR.is_dir():
    EYEPY_DATA_DIR.mkdir(parents=True)


def load(name):
    data_dir = EYEPY_DATA_DIR / name
    if not data_dir.is_dir():
        download_path = EYEPY_DATA_DIR / (name + ".zip")
        urllib.request.urlretrieve(SAMPLE_DATA[name][0], download_path)
        with zipfile.ZipFile(download_path, "r") as zip_ref:
            zip_ref.extractall(EYEPY_DATA_DIR)
        download_path.unlink()

    return SAMPLE_DATA[name][1](EYEPY_DATA_DIR / name / "metaclean.xml")



# -*- coding: utf-8 -*-
from datetime import datetime

from eyepy.io.utils import _clean_ascii, _date_from_seconds, _get_first


# "HSF-OCT-103"
def oct_base_spec():
    return {  # Version identifier: HSF-OCT-xxx
        "Version": ("12s", _clean_ascii),
        # Number of A-Scans in each B-Scan, i.e. B-Scan width in pixel
        "SizeX": ("I", _get_first),
        # Number of B-Scans in OCT scan
        "NumBScans": ("I", _get_first),
        # Number of samples in an A-Scan, i.e. B-Scan height in pixel
        "SizeZ": ("I", _get_first),
        # Width of a B-Scan pixel in mm
        "ScaleX": ("d", _get_first),
        # Distance between two adjacent B-Scans in mm
        "Distance": ("d", _get_first),
        # Height of a B-Scan pixel in mm
        "ScaleZ": ("d", _get_first),
        # Width of the SLO image in pixel
        "SizeXSlo": ("I", _get_first),
        # Height of the SLO image in pixel
        "SizeYSlo": ("I", _get_first),
        # Width of a pixel in the SLO image in mm
        "ScaleXSlo": ("d", _get_first),
        # Height of a pixel in the SLO image in mm
        "ScaleYSlo": ("d", _get_first),
        # Horizontal field size of the SLO image in dgr
        "FieldSizeSlo": ("I", _get_first),
        # Scan focus in dpt
        "ScanFocus": ("d", _get_first),
        # Examined eye (zero terminated string). "OS": Left eye; "OD": Right eye
        "ScanPosition": ("4s", _clean_ascii),
        # Examination time. The structure holds an unsigned 64-bit date and time
        # value. It is encoded as 100ns units since beginning of January 1, 1601
        "ExamTime": (
            "Q",
            lambda x: _date_from_seconds(x[0], datetime(1601, 1, 1), 1e-7),
        ),
        # Scan pattern type:
        #   0 = Unknown pattern,
        #   1 = Single line scan (one B-Scan only),
        #   2 = Circular scan (one B-Scan only),
        #   3 = Volume scan in ART mode,
        #   4 = Fast volume scan,
        #   5 = Radial scan (aka. star pattern)
        "ScanPattern": ("i", _get_first),
        # Size of the Header preceding each B-Scan in bytes
        "BScanHdrSize": ("I", _get_first),
        # Unique identifier of this OCT-scan (zero terminated string). This is
        # identical to the number <SerID> that is part of the file name.
        # Format: n[.m] n and m are numbers. The extension .m exists only for
        # ScanPattern 1 and 2. Examples: 2390, 3433.2
        "ID": ("16s", _clean_ascii),
        # Unique identifier of the reference OCT-scan (zero terminated string).
        # Format: see ID, This ID is only present if the OCT-scan is part of a
        # progression otherwise this string is empty. For the reference scan of
        # a progression ID and ReferenceID are identical.
        "ReferenceID": ("16s", _clean_ascii),
        # Internal patient ID used by HEYEX.
        "PID": ("I", _get_first),
        # User-defined patient ID (zero terminated string).
        "PatientID": ("24s", _clean_ascii),
        # Patient's date of birth
        "DOB": (
            "d",
            lambda x: _date_from_seconds(x[0], datetime(1899, 12, 30), 60 * 60 * 24),
        ),
        # Internal visit ID used by HEYEX.
        "VID": ("I", _get_first),
        # User-defined visit ID (zero terminated string). This ID can be defined
        # in the Comment-field of the Diagnosis-tab of the Examination Data
        # dialog box. The VisitID must be defined in the first row of the
        # comment field. It has to begin with an "#" and ends with any
        # white-space character. It can contain up to 23 alpha-numeric
        # characters (excluding the "#")
        "VisitID": ("24s", _clean_ascii),
        # Date the visit took place. Identical to the date of an examination tab
        # in HEYEX.
        "VisitDate": (
            "d",
            lambda x: _date_from_seconds(x[0], datetime(1899, 12, 30), 60 * 60 * 24),
        ),
        # Type of grid used to derive thickness data. 0 No thickness data
        # available, >0 Type of grid used to derive thickness values. Thickness
        # data is only available for ScanPattern 3 and 4.
        "GridType": ("I", _get_first),
        # File offset of the thickness data in the file. If GridType is 0,
        # GridOffset is 0.
        "GridOffset": ("I", _get_first),
        # Type of a 2nd grid used to derive a 2nd set of thickness data.
        "GridType1": ("I", _get_first),
        # File offset of the 2 nd thickness data set in the file.
        "GridOffset1": ("I", _get_first),
        # Internal progression ID (zero terminated string). All scans of the
        # same progression share this ID.
        "ProgID": ("34s", _clean_ascii),
        # Spare bytes for future use. Initialized to 0.
        "__empty": ("1790s", _get_first),
    }


# "HSF-BS-103"
def bscan_base_spec():
    return {  # Version identifier (zero terminated string).
        "Version": ("12s", _clean_ascii),
        # Size of the B-Scan header in bytes. It is identical to the same value
        # of the file header.
        "BScanHdrSize": ("i", _get_first),
        # X-Coordinate of the B-Scan's start point in mm.
        "StartX": ("d", _get_first),
        # Y-Coordinate of the B-Scan's start point in mm.
        "StartY": ("d", _get_first),
        # X-Coordinate of the B-Scan's end point in mm. For circle scans, this
        # is the X-Coordinate of the circle's center point.
        "EndX": ("d", _get_first),
        # Y-Coordinate of the B-Scan's end point in mm. For circle scans, this
        # is the Y-Coordinate of the circle's center point.
        "EndY": ("d", _get_first),
        # Number of segmentation vectors
        "NumSeg": ("I", _get_first),
        # Offset of the array of segmentation vectors relative to the beginning
        # of this B-Scan header.
        "OffSeg": ("I", _get_first),
        # Image quality measure. If this value does not exist, its value is set
        # to INVALID.
        "Quality": ("f", _get_first),
        # Horizontal shift (in # of A-Scans) of the classification band against
        # the segmentation lines (for circular scan only).
        "Shift": ("I", _get_first),
        # Intra volume transformation matrix. The values are only available for
        # volume and radial scans and if alignment is turned off, otherwise the
        # values are initialized to 0.
        "IVTrafo": ("ffffff", lambda x: x),
        # Spare bytes for future use.
        "__empty": ("168s", _get_first),
    }



# -*- coding: utf-8 -*-
import cmath
import logging

import numpy as np
from skimage import transform

from eyepy.utils.filters import radius_filtergrid, theta_filtergrid

logger = logging.getLogger(__name__)


def circle_mask(radius, mask_shape=None, smooth_edges=False):
    """Create a centered circular mask with given radius.

    Parameters
    ----------
    radius :
    mask_shape :
    smooth_edges :

    Returns
    -------
    """
    if mask_shape is None:
        mask_shape = (radius * 2, radius * 2)

    if smooth_edges:
        work_shape = (mask_shape[0] * 5, mask_shape[1] * 5)
        radius *= 5
    else:
        work_shape = mask_shape

    circle_mask = np.zeros(work_shape)
    circle_mask[
        radius_filtergrid(work_shape, quadrant_shift=False, normalize=False) < radius
    ] = 1

    return transform.resize(circle_mask, mask_shape)


def sector_masks(mask_shape, n_sectors=4, sector_rotations=0, smooth_edges=False):
    """Create masks for n radial sectors.

    By default the first sector is the first quadrant, and the remaining 3
    sectors are added counter clockwise.

    For a binary mask pixels can not belong to two mask without changing the
    sum over all masks. But for pixels at the sector edges it is not clear to
    which sector they belong and assigning them to two sectors partially might
    be desired. Hence if smooth_edges is True, we create 5 times bigger binary
    masks and then use the anti-aliasing from downscaling them to the desired
    shape to create a float mask.

    Parameters
    ----------
    mask_shape :
    n_sectors :
    sector_rotations :
    smooth_edges :

    Returns
    -------
    """
    if smooth_edges:
        work_shape = (mask_shape[0] * 5, mask_shape[1] * 5)
    else:
        work_shape = mask_shape

    theta = theta_filtergrid(work_shape, quadrant_shift=False)
    # Convert from angles in radian range [-pi, +pi] to degree range [0, 360]
    theta = theta / np.pi * 180
    theta[np.where(theta < 0)] += 360

    masks = []
    sector_size = 360 / n_sectors
    for i in range(n_sectors):
        sector_start = i * sector_size + sector_rotations
        sector_end = (i + 1) * sector_size + sector_rotations

        if sector_end >= 360 and sector_start >= 360:
            sector_end -= 360
            sector_start -= 360

        if sector_end >= 360:
            mask = np.ones(work_shape)
            mask[
                np.where(
                    np.logical_and(theta < sector_start, theta >= sector_end - 360)
                )
            ] = 0
        else:
            mask = np.zeros(work_shape)
            mask[
                np.where(np.logical_and(theta >= sector_start, theta < sector_end))
            ] = 1

        if smooth_edges:
            mask = transform.resize(mask, mask_shape)

        masks.append(mask)

    return masks


def create_region_shape_primitives(
    mask_shape,
    radii: list = (0.8, 1.8),
    n_sectors: list = (1, 4),
    rotation: list = (0, 45),
    center=None,
):
    """Create circles and lines indicating region boundaries of quantification
    masks. These can used for plotting the masks.

    Parameters
    ----------
    mask_shape :
    radii :
    n_sectors :
    rotation :
    center :

    Returns
    -------
    """
    if center is None:
        center = (mask_shape[0] / 2, mask_shape[0] / 2)

    primitives = {"circles": [], "lines": []}
    # Create circles
    for radius in radii:
        primitives["circles"].append({"center": center, "radius": radius})

    for i, (n_sec, rot, radius) in enumerate(zip(n_sectors, rotation, radii)):
        rot = rot / 360 * 2 * np.pi
        if not n_sec is None and n_sec != 1:
            for sec in range(n_sec):
                theta = 2 * np.pi / n_sec * sec + rot

                start = cmath.rect(radii[i - 1], theta)
                start = (start.real + center[0], start.imag + center[1])

                end = cmath.rect(radius, theta)
                end = (end.real + center[0], end.imag + center[1])

                primitives["lines"].append({"start": start, "end": end})

    return primitives


def create_region_masks(
    mask_shape,
    radii: list = (0.8, 1.8),
    n_sectors: list = (1, 4),
    rotation: list = (0, 45),
    center=None,
    smooth_edges=False,
    ring_sectors=True,
    add_circle_masks=False,
) -> list:
    """Create segmented circular region masks for quantification.

    First circular masks with the provided radii are generated. Then ring masks
    are created by subtracting the first circular mask from the second and so
    on.
    Ring masks are only created if they have a respective entry in n_sectors.
    If you want the complete ring, set the respective n_sectors entry to 1. You
    can skip a ring by setting the respective entry to None and you  can split
    the ring into n sectors by setting the respective entry to n.
    Setting a number in `rotation` rotates the respective rings sectors by n
    degree counterclockwise.

    Parameters
    ----------
    mask_shape :
    radii :
    n_sectors :
    rotation :
    center :
    smooth_edges :
    ring_sectors : If False, sectors are not applied to rings but to the complete circles
    add_circle_masks :

    Returns
    -------
    """

    # Create circles
    circles = []
    for radius in radii:
        circles.append(circle_mask(radius, mask_shape, smooth_edges))

    n_sectors = [
        n_sectors[i] if i < len(n_sectors) else None for i, _ in enumerate(radii[1:])
    ]
    rotation = [
        rotation[i] if i < len(rotation) else 0 for i, _ in enumerate(radii[1:])
    ]

    level_sector_parts = []
    for n_sec, rot in zip(n_sectors, rotation):
        if n_sec is not None:
            level_sector_parts.append(
                sector_masks(
                    mask_shape,
                    n_sectors=n_sec,
                    sector_rotations=rot,
                    smooth_edges=smooth_edges,
                )
            )

    if ring_sectors:
        rings = []
        for i, _ in enumerate(circles):
            if i + 1 >= len(circles):
                break
            elif not radii[i] < radii[i + 1]:
                break
            else:
                rings.append(-circles[i] + circles[i + 1])

        pairs = zip(rings, level_sector_parts)
    else:
        pairs = zip(circles, level_sector_parts)

    all_masks = []
    for cir, sectors in pairs:
        for sec in sectors:
            all_masks.append(cir * sec)

    if add_circle_masks:
        [all_masks.append(cir) for cir in circles]

    if center is not None:
        translation = transform.AffineTransform(
            translation=np.array(center) - np.array(mask_shape) / 2
        )
        all_masks = [transform.warp(mask, translation.inverse) for mask in all_masks]

    return all_masks



# -*- coding: utf-8 -*-
from eyepy.io.utils import (
    _get_date_from_xml,
    _get_datetime_from_xml,
    _get_first_as_float,
    _get_first_as_int,
    _get_first_as_str,
)


def oct_base_spec():
    return {
        # Version
        "Version": ("./SWVersion/Version", _get_first_as_str),
        # Number of A-Scans in each B-Scan, i.e. B-Scan width in pixel
        "SizeX": (
            ".//ImageType[Type='OCT']../" "OphthalmicAcquisitionContext/Width",
            _get_first_as_int,
        ),
        # Number of B-Scans in OCT scan
        "NumBScans": (".//ImageType[Type='OCT']..", len),
        # Number of samples in an A-Scan, i.e. B-Scan height in pixel
        "SizeZ": (
            ".//ImageType[Type='OCT']../" "OphthalmicAcquisitionContext/Height",
            _get_first_as_int,
        ),
        # Width of a B-Scan pixel in mm
        "ScaleX": (
            ".//ImageType[Type='OCT']../" "OphthalmicAcquisitionContext/ScaleX",
            _get_first_as_float,
        ),
        # Distance between two adjacent B-Scans in mm (Not given in the XMl we
        # could try to estimate it from the B-Scan positions)
        "Distance": ("", lambda x: None),
        # Height of a B-Scan pixel in mm (In the XML the scale is given for
        # every B-Scan and it is called ScaleY instead of ScaleZ)
        "ScaleZ": (
            ".//ImageType[Type='OCT']../" "OphthalmicAcquisitionContext/ScaleY",
            _get_first_as_float,
        ),
        # Width of the SLO image in pixel
        "SizeXSlo": (
            ".//ImageType[Type='LOCALIZER']../" "OphthalmicAcquisitionContext/Width",
            _get_first_as_int,
        ),
        # Height of the SLO image in pixel
        "SizeYSlo": (
            ".//ImageType[Type='LOCALIZER']../" "OphthalmicAcquisitionContext/Height",
            _get_first_as_int,
        ),
        # Width of a pixel in the SLO image in mm
        "ScaleXSlo": (
            ".//ImageType[Type='LOCALIZER']../" "OphthalmicAcquisitionContext/ScaleX",
            _get_first_as_float,
        ),
        # Height of a pixel in the SLO image in mm
        "ScaleYSlo": (
            ".//ImageType[Type='LOCALIZER']../" "OphthalmicAcquisitionContext/ScaleY",
            _get_first_as_float,
        ),
        # Horizontal field size of the SLO image in dgr
        "FieldSizeSlo": (
            ".//ImageType[Type='LOCALIZER']../" "OphthalmicAcquisitionContext/Angle",
            _get_first_as_int,
        ),
        # Scan focus in dpt
        "ScanFocus": (
            ".//ImageType[Type='LOCALIZER']../" "OphthalmicAcquisitionContext/Focus",
            _get_first_as_float,
        ),
        # Examined eye (zero terminated string). "OS": Left eye; "OD": Right eye
        "ScanPosition": (
            "./Patient/Study/Series/Laterality",
            lambda x: "OD" if x[0].text == "R" else "OS",
        ),
        # Examination time
        "ExamTime": ("./Patient/Study/Series", _get_datetime_from_xml),
        # Scan pattern type: (Not given in XML)
        #   0 = Unknown pattern,
        #   1 = Single line scan (one B-Scan only),
        #   2 = Circular scan (one B-Scan only),
        #   3 = Volume scan in ART mode,
        #   4 = Fast volume scan,
        #   5 = Radial scan (aka. star pattern)
        "ScanPattern": ("", lambda x: 0),
        # Unique identifier of this OCT-scan (zero terminated string). This is
        # identical to the number <SerID> that is part of the file name.
        # Format: n[.m] n and m are numbers. The extension .m exists only for
        # ScanPattern 1 and 2. Examples: 2390, 3433.2
        "ID": ("./Patient/Study/Series/ID", _get_first_as_int),
        # Unique identifier of the reference OCT-scan (zero terminated string).
        # Format: see ID, This ID is only present if the OCT-scan is part of a
        # progression otherwise this string is empty. For the reference scan of
        # a progression ID and ReferenceID are identical.
        "ReferenceID": ("./Patient/Study/Series/ReferenceSeries/ID", _get_first_as_int),
        # Internal patient ID used by HEYEX.
        "PID": ("./Patient/ID", _get_first_as_int),
        # User-defined patient ID (zero terminated string).
        "PatientID": ("./Patient/PatientID", _get_first_as_str),
        # Patient's date of birth
        "DOB": ("./Patient/Birthdate/Date", _get_date_from_xml),
        # Internal visit ID used by HEYEX.
        "VID": ("", lambda x: None),
        # User-defined visit ID (zero terminated string). This ID can be defined
        # in the Comment-field of the Diagnosis-tab of the Examination Data
        # dialog box. The VisitID must be defined in the first row of the
        # comment field. It has to begin with an "#" and ends with any
        # white-space character. It can contain up to 23 alpha-numeric
        # characters (excluding the "#")
        "VisitID": ("", lambda x: None),
        # Date the visit took place. Identical to the date of an examination tab
        # in HEYEX.
        "VisitDate": ("./Patient/Study/StudyDate/Date", _get_date_from_xml),
        # Type of grid used to derive thickness data. 0 No thickness data
        # available, >0 Type of grid used to derive thickness values. Thickness
        # data is only available for ScanPattern 3 and 4.
        "GridType": ("./Patient/Study/Series/ThicknessGrid/Type", _get_first_as_int),
        # Internal progression ID (zero terminated string). All scans of the
        # same progression share this ID.
        "ProgID": ("./Patient/Study/Series/ProgID", _get_first_as_str),
    }


def bscan_base_spec():
    return {
        # The same as the XML Version
        "Version": ("./SWVersion/Version", _get_first_as_str),
        # X-Coordinate of the B-Scan's start point in mm.
        "StartX": ("./OphthalmicAcquisitionContext/Start/Coord/X", _get_first_as_float),
        # Y-Coordinate of the B-Scan's start point in mm.
        "StartY": ("./OphthalmicAcquisitionContext/Start/Coord/Y", _get_first_as_float),
        # X-Coordinate of the B-Scan's end point in mm. For circle scans, this
        # is the X-Coordinate of the circle's center point.
        "EndX": ("./OphthalmicAcquisitionContext/End/Coord/X", _get_first_as_float),
        # Y-Coordinate of the B-Scan's end point in mm. For circle scans, this
        # is the Y-Coordinate of the circle's center point.
        "EndY": ("./OphthalmicAcquisitionContext/End/Coord/Y", _get_first_as_float),
        # Number of segmentation vectors
        "NumSeg": ("./Segmentation/NumSegmentations", _get_first_as_int),
        # Image quality measure. If this value does not exist, its value is set
        # to INVALID.
        "Quality": ("./OphthalmicAcquisitionContext/ImageQuality", _get_first_as_float),
        # Horizontal shift (in # of A-Scans) of the classification band against
        # the segmentation lines (for circular scan only).
        "Shift": ("", lambda x: None),
        # Intra volume transformation matrix. The values are only available for
        # volume and radial scans and if alignment is turned off, otherwise the
        # values are initialized to 0.
        "IVTrafo": ("", lambda x: None),
    }
